"""
Read/Write data on xml file for very simple database


Args:
    xmlFile (str | pathlib.Path): xml file name or Path object

Raises:
    ValueError: Raised when the xml file can not be created
        when first run and the file still don't exist

Examples:

    db = XmlDB("/dir/data.xml")
    db.initXmlFile('Data', ['Group1', 'Group2'])

    # if the record exist it data is update
    attrib = {'type': 'type1', 'other': 'one'} # can be empty
    db.add('Group1', 'id1' , a, 'Record 1')

    attrib = {'type': 'type2', 'other': 'two'} # can be empty
    db.add('Group2', 'id1' , a, 'Record 2')

    data = db.get('Group1', 'id1')
    db.remove('Group2', 'id1')


    # search is done by attributes inclusive all have to match
    # no key is needed this two are enough to pick only one
    # record
    attrib = {'type': 'type1', 'other': 'one'}
    data = db.search(attrib)

    In the example the record id is the same but group+id is unique
    and the get uses the group

::

 xml format:

    <?xml version="1.0" ?>
    <Data>
        <Group1>
            <Record type='type1' other='one' recordID='id1' recordType='python-type'>Record 1</Record>
        </Group1>
        <Group2>
            <Record type='type2' other='one' recordID='id1' recordType='python-type'>Record 2</Record>
        </Group2>
    </Data>

    recordType can be:

        - bool, byes, numbers: int, float, complex, strings
        - and dictionaries, lists and tuples of basic data types

    binary data that can converted to base64 and save as bytes
    can work the conversion from byres to binary is not
    manage by the class
"""

# XML0001


# pylint: disable=protected-access
# used for isinstance check


# standar imports
import ast
from pathlib import Path

# third-party imports
import lxml
import lxml.etree as ET


class XmlDB:
    """simple xml file database"""

    def __init__(self, xmlFile=None):

        self._xmlFile = None
        self._tree = None
        self._root = None
        self._status = None

        if xmlFile is not None:

            self.xmlFile = xmlFile

    def __bool__(self):

        return self._xmlFile is not None

    @property
    def root(self):
        """root of xml db"""
        return self._root

    @property
    def status(self):
        """Error message"""
        return self._status

    @property
    def xmlFileName(self):
        """xml file name string"""
        return str(self._xmlFile)

    @property
    def xmlDocPretty(self):
        """pretty printed xml document"""
        return ET.tostring(self._root, pretty_print=True).decode()

    @property
    def xmlFile(self):
        """xml file Path object"""
        return self._xmlFile

    @xmlFile.setter
    def xmlFile(self, xmlFile):
        """
        setup xml file paremeter can be a string or a Path
        """

        bError = False
        eMsg = ""

        if xmlFile is not None:

            if isinstance(xmlFile, Path):
                p = xmlFile
            else:
                p = Path(xmlFile)

            if not p.is_dir():
                if not p.is_file():
                    if p.parent.is_dir():
                        p.touch()

                if p.is_file():
                    self._xmlFile = p
                    if p.stat().st_size > 0:
                        self._openXmlFile()
                else:
                    eMsg = str(p) + " could not be created"
                    bError = True
            else:
                eMsg = str(p) + " is a directory is an error"
                bError = True
        else:
            eMsg = "File parameter is None"
            bError = True

        if bError:
            self._xmlFile = None
            self._status = eMsg
            raise ValueError(eMsg)

    def initXmlFile(self, root, groups, force=False):
        """init the xml file if empty force=True do clear"""

        if self._xmlFile is not None:

            size = self._xmlFile.stat().st_size

            if not size or force:
                r = ET.Element(root)

                for tag in groups:
                    ET.SubElement(r, tag)

                e = ET.SubElement(r, Key.InitializationValues)
                cfg = ET.SubElement(e, 'FileRoot')
                cfg.text = root
                cfg = ET.SubElement(e, 'Groups')
                cfg.text = str(groups)

                tree = ET.ElementTree(r)

                tree.write(self.xmlFileName)

                self._status = "File initialized."
            else:
                self._status = "File already initialized.\nUse force=True to reinitialized (WARNING! all records are purged)"

            self._openXmlFile()

        else:

            raise RuntimeError("XML file not defined.")

    def _openXmlFile(self):
        """open xml file"""

        if self._xmlFile is not None:

            #try: lxml.etree.XMLSyntaxError
            #TypeError
            try:
                self._tree = ET.parse(self.xmlFileName)
                self._root = self._tree.getroot()
            except lxml.etree.XMLSyntaxError as error:
                self._status = "Error on file {} - {} ".format(
                    self.xmlFileName,
                    error.msg
                )
            except TypeError as error:
                self._status = str(error.args)

    def add(self, group, recordID, attrib, data):
        """add record to xml group"""

        r = self._root
        attr = attrib

        if not isinstance(recordID, str):
            strID = str(recordID)
        else:
            strID = recordID

        attr[Key.recordID] = strID
        attr[Key.recordType] = type(data).__name__

        ge = r.find('.//' + group) # group to actualize

        if isinstance(ge, ET._Element):

            # search record
            find = ET.XPath('.//' + group + '/' + 'Record[@recordID="' + str(recordID) + '"]')
            try:
                e = find(r)[0]
            except IndexError:
                e = None

            if isinstance(e, ET._Element):
                e.text = str(data)
            else:
                e = ET.SubElement(ge, "Record")
                for key in attr:
                    e.set(key, attr[key])
                e.text = str(data)

            self._tree.write(self.xmlFileName)

            self._status = "Record added"
        else:
            self._status = "Group not found: group - {}".format(
                group
            )

    def get(self, group, recordID):
        """get record"""

        if not isinstance(recordID, str):
            strID = str(recordID)
        else:
            strID = recordID

        value = None

        r = self._root

        # search record
        find = ET.XPath('.//' + group + '/' + 'Record[@recordID="' + strID + '"]')

        try:
            e = find(r)[0]
        except IndexError:
            e = None

        if isinstance(e, ET._Element):

            if e.attrib[Key.recordType] == "str":
                value = e.text
            else:
                value = ast.literal_eval(e.text)

            self._status = "Get record successful"

        else:
            self._status = "Record not found: group - {} record id - {}".format(
                group,
                strID
            )

        return value

    def getGroupRoot(self, group):
        """returns the root of a group"""

        r = self._root
        find = ET.XPath('.//' + group)

        try:
            e = find(r)[0]
        except IndexError:
            e = None

        return e

    def remove(self, group, recordID):
        """add record to xml group"""

        if not isinstance(recordID, str):
            strID = str(recordID)
        else:
            strID = recordID

        r = self._root

        # search record
        find = ET.XPath('.//' + group + '/' + 'Record[@id="' + strID + '"]')
        try:
            e = find(r)[0]
        except IndexError:
            e = None

        if isinstance(e, ET._Element):
            # get parent of record then remove from parent
            p = e.getparent()
            p.remove(e)

            self._tree.write(self.xmlFileName)

            self._status = "Remove successful"
            return True
        else:
            self._status = "Record to remove not found: group - {} record id - {}".format(
                group,
                strID
            )
        return False

    def search(self, attrib):
        """add record to xml group
        [contains(@name, "Jack Reacher") and contains(@date, "2016") and contains(@type, "Movie")]
        """

        value = None
        if not isinstance(attrib, dict):
            return None

        r = self._root
        sa = _attribToString(attrib)
        find = ET.XPath('.//Record' + sa)

        try:
            s = find(r)[0]
        except IndexError:
            s = None
        except TypeError:
            s = None

        if isinstance(s, ET._Element):
            if s.attrib[Key.recordType] == "str":
                value = s.text
            else:
                value = ast.literal_eval(s.text)
            self._status = "Last search successful"
        else:
            self._status = "Last search failed: " + str(attrib)


        return value


class Key:
    """Dictionary keys"""

    recordID = "recordID"
    recordType = "recordType"
    InitializationValues = "InitializationValues"
    Movie = "Movie"
    TV = "TV"

def _attribToString(attrib):
    """
    convert attrib dictionnary to string suitable for XPath search

    {'name': 'Reacher', 'date': '2016', 'type': 'Movie'}
    [contains(@name, "Reacher") and contains(@type, "Movie")]
    """

    s = "["

    lFirst = True

    for k in attrib:
        if lFirst:
            s = s + "contains(@" + k + ', "' + attrib[k] + '")'
            lFirst = False
        else:
            s = s + " and contains(@" + k + ', "' + attrib[k] + '")'

    s = s + "]"

    return s



"""
pretty print tree
"""

from pathlib import Path

from natsort import natsorted, ns


class Prefix:

    filenameMiddle = "├──"
    filenameLast = "└──"
    parentMiddle = "    "
    parentLast = "│   "

    # filenameMiddle = "├─"
    # filenameLast = "└─"
    # parentMiddle = "  "
    # parentLast = "│ "


class DisplayPath:
    """DisplayPath"""

    def __init__(self, path, parentPath, isLast):
        self.path = Path(str(path))
        self.parentPath = parentPath
        self.isLast = isLast
        if self.parentPath:
            self.depth = self.parentPath.depth + 1
        else:
            self.depth = 0

    @classmethod
    def makeTree(cls, root, parent=None, isLast=False, criteria=None, fileList=None):
        """makeTree"""

        root = Path(str(root))
        criteria = criteria or cls._defaultCriteria

        displayableRoot = cls(root, parent, isLast)
        yield displayableRoot

        childNames = []
        if fileList is not None:
            for oFile in fileList:
                childNames.append(oFile.name)
            children = natsorted(fileList, alg=ns.PATH)
        else:
            children = natsorted(
                list(path for path in root.iterdir() if criteria(path)), alg=ns.PATH,
            )
        count = 1
        for path in children:
            isLast = count == len(children)
            if path.is_dir():
                yield from cls.makeTree(
                    path, parent=displayableRoot, isLast=isLast, criteria=criteria
                )
            else:
                yield cls(path, displayableRoot, isLast)
            count += 1

    @classmethod
    def _defaultCriteria(cls, path):  # pylint: disable=unused-argument
        return True

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + "/"
        return self.path.name

    def displayable(self):
        """displayable"""

        if self.parentPath is None:
            return self.displayname

        _filename_prefix = Prefix.filenameLast if self.isLast else Prefix.filenameMiddle

        parts = ["{!s} {!s}".format(_filename_prefix, self.displayname)]

        parent = self.parentPath
        while parent and parent.parentPath is not None:
            parts.append(Prefix.parentMiddle if parent.isLast else Prefix.parentLast)
            parent = parent.parentPath

        return "".join(reversed(parts))



"""
Decorators found on internet
"""


import functools


def callCounter(func):
    """function call counter"""

    @functools.wraps(func)
    def helper(*args, **kwargs):
        helper.calls += 1
        return func(*args, **kwargs)
    helper.calls = 0

    return helper

def staticVars(**kwargs):
    """Add static variables to function"""

    def decorate(func):
        """Add attributes to func"""
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate



"""xml utils"""

import xml.dom.minidom as DOM

def xmlPrettyPrint(xmlString, indent=None):
    """
    Convenience function that receives a xml string
    and returns it pretty printed

    Args:
        xmlString (bytes): a bytes xml string
        indent (str, optional): space padding string. Defaults to None.

    Returns:
        str: pretty printed formmated xml string
    """

    i = "    "
    if indent is not None:
        i = indent

    xDoc = DOM.parseString(xmlString)

    if xDoc:
        xPretty = xDoc.toprettyxml(indent=i)

    return xPretty



"""
function for QMessageBox Yes|No needed for change
font on the fly and localization

Args:
    self (QWidget): parent QWidget
    tittle (str): tittle for QMessageBox
    text (str): text for QMessageBox
    icon (QIcon): standard icons for QMessageBox
        question mark, warning ...
Returns:
    QMessageBox.button: standard QMessageBox button

"""

from PySide2.QtWidgets import QMessageBox


def messageBox(self, title, text, icon=QMessageBox.Information):
    """
    Working on generic message box
    """

    m = QMessageBox(self)
    m.setWindowTitle(title)
    m.setText(text)
    m.setIcon(icon)
    # yesButton = m.addButton('Yes', QMessageBox.ButtonRole.YesRole)
    # noButton = m.addButton('No', QMessageBox.ButtonRole.NoRole)
    m.setDefaultButton(QMessageBox.Ok)
    m.setFont(self.font())
    m.exec_()

    return QMessageBox.Ok


def messageBoxYesNo(self, title, text, icon):
    """
    Yes | No message box
    """

    m = QMessageBox(self)
    m.setWindowTitle(title)
    m.setText(text)
    m.setIcon(icon)
    yesButton = m.addButton("Yes", QMessageBox.ButtonRole.YesRole)
    noButton = m.addButton("No", QMessageBox.ButtonRole.NoRole)
    m.setDefaultButton(noButton)
    m.setFont(self.font())
    m.exec_()

    if m.clickedButton() == yesButton:
        return QMessageBox.Yes

    return QMessageBox.No



"""
qtUtils:

utility functions that use PySide2
"""

from PySide2.QtGui import QColor

from vsutillib.macos import isMacDarkMode

from .SvgColor import SvgColor


def checkColor(color, isDarkMode=False):
    """
    checkColor change color according to dark or light mode

    Args:
        color (SvgColor): color code
        isDarkMode (bool, optional): True if using dark mode on Windows. Defaults to False.

    Returns:
        [type]: [description]
    """

    if color is None:

        if isMacDarkMode() or isDarkMode:
            color = SvgColor.white
        else:
            color = SvgColor.black

    elif isMacDarkMode():

        if color == SvgColor.red:
            color = SvgColor.magenta
        elif color == SvgColor.darkgreen:
            color = SvgColor.green
        elif color == SvgColor.blue:
            color = SvgColor.cyan

    elif not isDarkMode:

        if color == SvgColor.cyan:
            color = SvgColor.dodgerblue
        elif color == QColor(42, 130, 218):
            color = SvgColor.dodgerblue

    return color


class LineOutput:

    Color = "color"
    ReplaceLine = "replaceLine"
    AppendLine = "appendLine"
    AppendEnd = "appendEnd"
    LogOnly = "logOnly"



"""
url search class
"""

import urllib

class UrlRequest():
    """
    Issue url request response is set in
    status and message properties
    POST if data is not None
    GET is data is None

    Args:
        url (:obj:`str`, optional): URL for request
        data (:obj:`dict`, optional): dictionary with data for GET
        headers (:obj:`dict`, optional): dictionary with any needed headers

    Attributes:
        status (int): status code from request
        message (str): message response from request
    """

    def __init__(self, url=None, data=None, headers=None):

        self.url = url
        self.data = data
        self.headers = headers

        self.status = None
        self.message = None

        self._urlSearch()

    def request(self, url=None, data=None, headers=None):
        """
        Issue a new request for url

        Args:
            url (str): URL for request
            data (:obj:`dict`, optional): dictionary with data for GET
            headers (:obj:`dict`, optional): dictionary with any needed headers

        """
        self.url = url
        self.status = data
        self.message = headers

        self.status = None
        self.message = None

        self._urlSearch()

    def _urlSearch(self):
        """
        request url and update
        status and message properties
        """

        if self.url is not None:

            if self.data is not None:
                request = urllib.request.Request(self.url, self.data)
            else:
                request = urllib.request.Request(self.url)

            if self.headers is not None:
                for key in self.headers:
                    request.add_header(key, self.headers[key])

            result = urllib.request.urlopen(request)

            if result:

                self.status = result.status
                self.message = result.read()



"""

The MIT License (MIT)

Copyright (c) 2011 Morgan Leborgne

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


QProgressIndicator

Returns:
    [type]: [description]
"""

import sys

from PySide2.QtCore import Qt, QSize, Signal, Slot
from PySide2.QtGui import QPainter, QColor
from PySide2.QtWidgets import QWidget, QApplication, QSizePolicy

from .SvgColor import SvgColor


class QProgressIndicator(QWidget):
    """
    ProgressIndicator show progress animation
    """

    startAnimationSignal = Signal()
    stopAnimationSignal = Signal()


    def __init__(self, parent):
        # Call parent class constructor first
        super(QProgressIndicator, self).__init__(parent)

        # Initialize instance variables
        self.angle = 0
        self.timerId = -1

        self.__delay = 40
        self.__displayedWhenStopped = False
        self.__color = SvgColor.black

        # Set size and focus policy
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.setFocusPolicy(Qt.NoFocus)

        self.startAnimationSignal.connect(self.startAnimation)
        self.stopAnimationSignal.connect(self.stopAnimation)

        # Show the widget
        self.show()

    @property
    def color(self):
        return self.__color

    @color.setter
    def color(self, value):
        self.__color = value
        self.update()

    @property
    def delay(self):
        return self.__delay

    @delay.setter
    def delay(self, value):
        self.__delay = value
        if self.timerId != -1:
            self.killTimer(self.timerId)
            self.timerId = self.startTimer(self.__delay)

    @property
    def displayedWhenStopped(self):
        return self.__displayedWhenStopped

    @displayedWhenStopped.setter
    def displayedWhenStopped(self, value):
        self.__displayedWhenStopped = value
        self.update()

    def heightForWidth(self, w):
        return w

    def sizeHint(self):
        return QSize(30, 30)

    def timerEvent(self, event):  # pylint: disable=unused-argument
        self.angle = (self.angle + 30) % 360
        self.update()

    def paintEvent(self, event):  # pylint: disable=unused-argument
        if (not self.displayedWhenStopped) and (not self.isAnimated()):
            return

        width = min(self.width(), self.height())

        p = QPainter(self)
        p.setRenderHint(QPainter.Antialiasing)

        outerRadius = (width - 1) * 0.5
        innerRadius = (width - 1) * 0.5 * 0.38

        capsuleHeight = outerRadius - innerRadius
        capsuleWidth = capsuleHeight * 0.23 if (width > 32) else capsuleHeight * 0.35
        capsuleRadius = capsuleWidth / 2

        for i in range(0, 12):
            color = QColor(self.color)

            color.setAlphaF(1.0 - (i / 12.0))

            p.setPen(Qt.NoPen)
            p.setBrush(color)
            p.save()
            p.translate(self.rect().center())
            p.rotate(self.angle - (i * 30.0))
            p.drawRoundedRect(
                -capsuleWidth * 0.5,
                -(innerRadius + capsuleHeight),
                capsuleWidth,
                capsuleHeight,
                capsuleRadius,
                capsuleRadius,
            )
            p.restore()

    def isAnimated(self):
        return self.timerId != -1

    @Slot()
    def startAnimation(self):
        self.angle = 0
        if self.timerId == -1:
            self.timerId = self.startTimer(self.delay)

    @Slot()
    def stopAnimation(self):
        if self.timerId != -1:
            self.killTimer(self.timerId)

        self.angle = 0
        self.timerId = -1
        self.update()


if __name__ == "__main__":

    from PySide2.QtWidgets import QGridLayout, QMainWindow, QPushButton

    class MainWindow(QMainWindow):
        """Test the progress bars"""

        def __init__(self, *args, **kwargs):
            super(MainWindow, self).__init__(*args, **kwargs)

            l = QGridLayout()

            self.pi = QProgressIndicator(self)
            self.pi.displayedWhenStopped = True

            a = QPushButton("Animate")
            a.pressed.connect(self.animate)
            s = QPushButton("Stop Animation")
            s.pressed.connect(self.stopAnimation)

            l.addWidget(self.pi, 0, 0, Qt.AlignCenter)
            l.addWidget(a, 1, 0)
            l.addWidget(s, 2, 0)

            w = QWidget()
            w.setLayout(l)

            self.setCentralWidget(w)

            self.show()

        def animate(self):
            self.pi.startAnimation()

        def stopAnimation(self):
            self.pi.stopAnimation()

    # pylint: disable=invalid-name
    app = QApplication([])
    window = MainWindow()
    sys.exit(app.exec_())



"""
class emit signal that can be received by many Slots
"""

from PySide2.QtCore import QObject, Signal

class QGroupSignal(QObject):
    """
    SetLanguage class to save and trigger multiple slots

    Args:
        QObject (QObject): base class in order to work with Signals
    """

    groupSignal = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.parent = parent
        self.functionSlots = []

    def addSlot(self, function):

        self.groupSignal.connect(function)
        self.functionSlots.append(function)

    def emitSignal(self):

        self.groupSignal.emit()



"""
Get file structure information from media file
"""

# MFI0001

import base64
import logging

from pymediainfo import MediaInfo

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class MediaFileInfo(object):
    """
    Media file properties
    This class main function is to verify that the structure of the files
    is the same:

        Same order of tracks
        Same language if it applies

    Args:

        strMediaFile (str): name with fullpath of source file

    Attributes:

        lstMediaTracks (MediaTrackInfo): list of media tracks
            found
    """

    # log state
    __log = False

    @classmethod
    def classLog(cls, setLogging=None):
        """
        get/set logging at class level
        every class instance will log
        unless overwritten

        Args:
            setLogging (bool):
                - True class will log
                - False turn off logging
                - None returns current Value

        Returns:
            bool:

            returns the current value set
        """

        if setLogging is not None:
            if isinstance(setLogging, bool):
                cls.__log = setLogging

        return cls.__log

    def __init__(self, strMediaFile, log=None):

        self.fileName = strMediaFile
        self.mediaInfo = None
        self.codec = ""
        self.format = ""
        self.title = ""
        self.lstMediaTracks = []
        self.__log = None
        self.log = log
        self.totalTracks = {"Video": 0, "Audio": 0, "Text": 0}
        self._initHelper()

    def _initHelper(self):

        try:
            fileMediaInfo = MediaInfo.parse(self.fileName)
            self.mediaInfo = fileMediaInfo
        except OSError:
            raise OSError("MediaInfo not found.")

        for track in fileMediaInfo.tracks:
            if track.track_type == "General":
                self.codec = track.codec
                self.format = track.format
                if track.title is not None:
                    self.title = track.title
                    try:
                        self.title = base64.b64decode(track.title).decode("UTF-8")
                    except: # pylint: disable=bare-except
                        pass
                    self.title = self.title.strip()
            if track.track_type in ("Video", "Audio", "Text"):
                self.totalTracks[track.track_type] += 1
                self.lstMediaTracks.append(
                    MediaTrackInfo(track.streamorder, track.track_type,
                                   track.language, track.default, track.forced,
                                   track.title, track.codec, track.format))
    def __len__(self):
        return len(self.lstMediaTracks) if self.lstMediaTracks else 0

    def __eq__(self, objOther):

        bReturn = True

        if self.log:
            MODULELOG.debug(
                "MFI0001: Structure equality test between [%s] and [%s]",
                self.fileName, objOther.fileName)
            MODULELOG.debug("MFI0002: FORMAT: %s", self.format)

        if self.codec != objOther.codec:
            if self.log:
                MODULELOG.debug("MFI0003: Codec mismatched %s - %s", self.codec,
                                objOther.codec)
            bReturn = False
        elif len(self) != len(objOther):
            if self.log:
                MODULELOG.debug("MFI0004: Number of tracks mismatched %s - %s",
                                len(self), len(objOther))
            bReturn = False
        elif len(self) == len(objOther):
            for a, b in zip(self.lstMediaTracks, objOther.lstMediaTracks):
                if a.streamorder != b.streamorder:
                    if self.log:
                        MODULELOG.debug(
                            "MFI0005:  Stream order mismatched %s - %s",
                            a.streamorder, b.streamorder)
                    bReturn = False
                elif a.track_type != b.track_type:
                    if self.log:
                        MODULELOG.debug(
                            "MFI006: Stream type mismatched %s - %s",
                            a.track_type, b.track_type)
                    bReturn = False
                elif (a.language != b.language) and (a.track_type != "Video"):
                    if self.log:
                        MODULELOG.debug(
                            "MFI0007: Stream language mismatched %s - %s",
                            a.language, b.language)
                    if self.format == 'AVI':
                        # Ignore language for AVI container
                        if self.log:
                            MODULELOG.debug(
                                "MFI0008: AVI container ignore language mismatched %s - %s",
                                a.track_type, b.track_type)
                    else:
                        bReturn = False
                elif (a.codec != b.codec) and (a.track_type != "Audio"):
                    if self.log:
                        MODULELOG.debug("MFI0009: Codec mismatched %s - %s",
                                        a.codec, b.codec)
                    if self.format == 'AVI':
                        # Ignore language for AVI container
                        if self.log:
                            MODULELOG.debug(
                                "MFI0010: AVI container ignore codec mismatched %s - %s",
                                a.codec, b.codec)
                    else:
                        bReturn = False
                elif a.format != b.format:
                    if self.log:
                        MODULELOG.debug(
                            "MFI0011: Stream format mismatched %s - %s",
                            a.format, b.format)
                    bReturn = False

        if self.log:
            if bReturn:
                MODULELOG.debug("MFI0012: Structure found ok.", )
            else:
                MODULELOG.debug("MFI0013: Structure not ok.", )

        return bReturn

    def __str__(self):

        tmpStr = "File Name: {}\nFile Format: -{}-\n\n".format(
            self.fileName, self.format)
        tmpNum = 1

        for track in self.lstMediaTracks:
            tmpStr += "Track: {}\n".format(tmpNum)
            tmpStr += "Order: {} - {}\n".format(track.streamorder,
                                                track.track_type)
            tmpStr += "Codec: {}\n".format(track.codec)
            tmpStr += "Language: {}\n".format(track.language)
            tmpStr += "Format: {}".format(track.format)
            tmpNum += 1

        return tmpStr

    @property
    def log(self):
        """
        class property can be used to override the class global
        logging setting

        Returns:
            bool:

            True if logging is enable False otherwise
        """
        if self.__log is not None:
            return self.__log

        return MediaFileInfo.classLog()

    @log.setter
    def log(self, value):
        """set instance log variable"""
        if isinstance(value, bool) or value is None:
            self.__log = value


class MediaTrackInfo:
    r"""
    Convenience class used by MediaFileInfo_
    contains the media track properties

    Args:
        streamorder (:obj:`str`, optional): order of track
        track_type (:obj:`str`, optional): type of track
        language (:obj:`str`, optional): language of track
        default (:obj:`str`, optional): is this a default track
        forced (:obj:`str`, optional): is this a forced track
        title (:obj:`str`, optional): title of track
        codec (:obj:`str`, optional): codec used on track
        format\_ (:obj:`str`, optional): format of track

    Attributes:
        streamorder (str): Stream order
        track_type (str): Track Type
        language (str): Language of track
        default (str): Is a default track
            'Yes' it is 'No' otherwise
        forced (str): Is a forced track
            'Yes' it is 'No' otherwise
        title (str): Track tittle
        codec (str): Codec used
        format (str): Track format
    """

    def __init__(self,
                 streamorder=None,
                 track_type=None,
                 language=None,
                 default=None,
                 forced=None,
                 title=None,
                 codec=None,
                 format_=None):

        self.streamorder = streamorder
        self.track_type = track_type  # pylint: disable=C0103
        self.language = language
        self.default = default
        self.forced = forced
        self.title = title
        self.codec = codec
        self.format = format_

    def __str__(self):
        return "Stream Order: " + str(self.streamorder) \
            + "\nTrack Type: " + str(self.track_type) \
            + "\nLanguage: " + str(self.language) \
            + "\nDefault Track : " + str(self.default) \
            + "\nForced Track: " + str(self.forced) \
            + "\nTrack Title: " + str(self.title) \
            + "\nCodec: " + str(self.codec) \
            + "\nFormat: " + str(self.format) \
            + "\n"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""setup file to build python distributions"""

import io
import os
import pathlib
import shutil

from distutils.command.clean import clean
from distutils.command.install import install
from pathlib import Path
from setuptools import setup

from vsutillib import config

ROOT = pathlib.Path(__file__).parent.resolve()


class MyInstall(install):
    """install subclass"""

    # Calls the default run command, then deletes the build area
    # (equivalent to "setup clean --all").
    def run(self):
        print("what to clean {}".format(self.distribution))
        install.run(self)
        c = clean(self.distribution)
        c.all = True
        c.finalize_options()
        c.run()


def removeTmpDirs():
    """
    delete build directory setup was including files from other builds
    """
    p = Path(".")
    eggDirs = [x for x in p.glob("*.egg-info") if x.is_dir()]
    eggDirs.append(Path("build"))

    for d in eggDirs:
        if d.is_dir():
            shutil.rmtree(d)


def readme():
    """get README.rst"""

    try:
        with io.open(os.path.join(ROOT, "README.rst"), encoding="utf-8") as f:
            long_description = "\n" + f.read()
    except FileNotFoundError:
        long_description = config.DESCRIPTION
    return long_description

removeTmpDirs()

setup(
    name=config.NAME,  # Required
    version=config.VERSION,  # Required
    description=config.DESCRIPTION,  # Required
    long_description=readme(),  # Optional
    author=config.AUTHOR,  # Optional
    author_email=config.EMAIL,  # Optional
    url=config.PYPI,
    license="MIT",
    classifiers=[  # Optional
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: MacOS X",
        "Environment :: Win32 (MS Windows)",
        "Environment :: X11 Applications :: Qt",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Multimedia :: Video",
        "Topic :: Software Development :: Libraries",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        # Operating Systems
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: POSIX :: Linux",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.8",
        # Implementation
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    keywords=config.KEYWORDS,  # Optional
    packages=config.PACKAGES,
    install_requires=config.REQUIRED,
    python_requires=config.PYTHONVERSIONS,
    include_package_data=True,
    project_urls=config.PROJECTURLS,
)

removeTmpDirs()



"""
sql related classes
"""

import re
import sqlite3

from sqlite3 import Error as SQLiteError


class SqlDb:
    """
    class to manage connection to a sqllite database

    Raises:
        ValueError: when parametes don't match
    """

    def __init__(self, dbFile=None, autoCommit=False):

        self.__dbFile = dbFile
        self.__conn = None
        self.__lastError = None
        self.__autoCommit = autoCommit

        if self.__dbFile is not None:
            self.connect(dbFile)

    def __bool__(self):
        return bool(self.__conn)

    @property
    def connection(self):
        return self.__conn

    @property
    def error(self):
        return self.__lastError

    def close(self):
        self.connection.close()

    def connect(self, database, autoCommit=False):
        """
        connect connects to sqlite database

        Args:
            database (str, optional): database file. Defaults to None.
            autoCommit (bool, optional): execute commit after detection of
                SQL statement. Defaults to False.
        """

        rc = False

        self.__autoCommit = autoCommit
        self.__conn = None
        self.__lastError = None

        if isinstance(database, str):
            dbFile = database
        elif self.__dbFile is not None:
            dbFile = self.__dbFile

        if dbFile is not None:
            try:
                self.__conn = sqlite3.connect(database)
                rc = True
            except SQLiteError as e:
                self.__lastError = "SQLiteError: {}".format(e)

        return rc

    def sqlExecute(self, sqlStatement, *args, test=True):
        """
        sqlExecute execute sql statement

        Args:
            sqlStatement (str): sql statement to execute
            test (bool, optional): verify parameters pass to the method.
                Defaults to True.

        Raises:
            ValueError: raises error if the number of fields and values does not match

        Returns:
            sqlite3.cursor: cursor to point to the result of some operations
        """

        if test:
            regEx = re.compile(r"\?")
            regSelectEx = re.compile("SELECT")
            numExpectedVariables = 0
            numVariables = len(args)

            if match := regEx.findall(sqlStatement):
                numExpectedVariables = len(match)

            if numExpectedVariables != numVariables:
                if numExpectedVariables == 0:
                    msg = "Unexpected variables - {}".format(args)
                elif numVariables == 0:
                    msg = "Missing variables expected {}".format(numExpectedVariables)
                else:
                    msg = "Expected variables was {} received {}".format(
                        numExpectedVariables, numVariables
                    )
                raise ValueError(msg)

        cursor = None
        self.__lastError = None

        if match := regSelectEx.match(sqlStatement.strip().upper()):
            try:
                if len(args) > 0:
                    cursor = self.connection.execute(sqlStatement, args)
                else:
                    cursor = self.connection.execute(sqlStatement)
            except SQLiteError as e:
                self.__lastError = "SQLiteError: {}".format(e)
        else:
            try:
                with self.connection as conn:
                    if len(args) > 0:
                        cursor = conn.execute(sqlStatement, args)
                    else:
                        cursor = conn.execute(sqlStatement)
            except sqlite3.IntegrityError:
                self.__lastError = "Could not execute {}".format(sqlStatement)
            except SQLiteError as e:
                self.__lastError = "SQLiteError: {}".format(e)

        return cursor

    def sqlExecuteOriginal(self, sqlStatement, *args, test=True):
        """
        sqlExecute execute sql statement

        Args:
            sqlStatement (str): sql statement to execute
            test (bool, optional): verify parameters pass to the method.
                Defaults to True.

        Raises:
            ValueError: raises error if the number of fields and values does not match

        Returns:
            sqlite3.cursor: cursor to point to the result of some operations
        """

        if test:
            regEx = re.compile(r"\?")
            regDeleteEx = re.compile("DELETE")
            regInsertEx = re.compile("INSERT")
            regUpdateEx = re.compile("UPDATE")
            numExpectedVariables = 0
            numVariables = len(args)

            if match := regEx.findall(sqlStatement):
                numExpectedVariables = len(match)

            if numExpectedVariables != numVariables:
                if numExpectedVariables == 0:
                    msg = "Unexpected variables - {}".format(args)
                elif numVariables == 0:
                    msg = "Missing variables expected {}".format(numExpectedVariables)
                else:
                    msg = "Expected variables was {} received {}".format(
                        numExpectedVariables, numVariables
                    )
                raise ValueError(msg)

        cursor = None
        self.__lastError = None

        try:
            if len(args) > 0:
                cursor = self.connection.execute(sqlStatement, args)
            else:
                cursor = self.connection.execute(sqlStatement)
            if match := regDeleteEx.match(sqlStatement.strip().upper()):
                print("Delete commit")
                if self.__autoCommit:
                    print("select.connection.commit()")
            if match := regInsertEx.match(sqlStatement.strip().upper()):
                print("Insert commit")
                if self.__autoCommit:
                    print("select.connection.commit()")
            if match := regUpdateEx.match(sqlStatement.strip().upper()):
                print("Update commit")
                if self.__autoCommit:
                    print("select.connection.commit()")
        except SQLiteError as e:
            self.__lastError = "SQLiteError: {}".format(e)

        return cursor

    def transaction(self):
        sqlBeginTransaction = "BEGIN TRANSACTION;"
        cursor = self.connection.execute(sqlBeginTransaction)
        if cursor:
            return True
        return False

    def rollback(self):
        sqlRollback = "ROLLBACK;"
        cursor = self.connection.execute(sqlRollback)
        if cursor:
            return True
        return False

    def commit(self):
        sqlCommit = "COMMIT;"
        cursor = self.connection.execute(sqlCommit)
        if cursor:
            return True
        return False

    def tableExists(self, dbTable):
        """
        verify if table exists in the database
        """

        sqlTableCheck = """
            SELECT count(name)
              FROM sqlite_master
              WHERE type='table' AND name=?
            """

        cursor = self.connection.cursor()
        cursor.execute(sqlTableCheck, (dbTable, ))

        if cursor:
            row = cursor.fetchone()
            if row:
                return bool(row[0])

        return False



"""VS module names"""

from .mkvattachments import MKVAttachment, MKVAttachments
from .mkvclassutil import SourceFile, SourceFiles
from .mkvcommand import MKVCommand
#from .mkvcommandNew import MKVCommandNew
from .mkvcommandparser import MKVCommandParser, MKVParseKey
from .verifystructure import VerifyStructure
from .verifycommand import VerifyMKVCommand



"""
 Parse the attachments from mkvmerge command line
 and gather all that is needed to add attachments

 Global:
  Apply the same attachments to all files.  If the application does
  not find enough directories in the parent directory of the supplied
  one or there are more than one directory for the attachments in
  the command line.

 Per File:
  If the application finds the same number of directories as
  the number of input files in the root of the supplied it will
  assumed is directory for each file.  The directories can be empty
  meaning corresponding file has no attachments.
"""

import re
import shlex

from pathlib import Path

from ..mkvutils import unQuote


class MKVAttachment:  # pylint: disable=too-few-public-methods
    """
    Class to save attachment information

    """

    def __init__(self, attachment, span=None, matchString=None):

        self.name = None
        self.mimeType = None
        self.fileName = None

        if isinstance(attachment, tuple):
            self.name = attachment[0]
            self.mimeType = attachment[1]

            f = unQuote(attachment[2])
            p = Path(f)

            try:
                test = p.is_file()
            except OSError:
                self.fileName = None
            else:
                if test:
                    self.fileName = p

        self.span = span
        self.matchString = matchString

    def __str__(self):

        return self.matchString


class MKVAttachments:
    """
     Get attachment files

     Case 1: More than directories with attachments
        Return MKVCommandParser.attachmentsString for every Source

     Case 2: One directory with attachments
        - files in directory differ from total attachments on CL
          return MKVCommandParser.attachmentsString for every Source
        - files in directory equals total attachments on CL
          read directories on parent directory
          * total direcoties equals total source read directories
            and return files read to corresponding source. Directories
            can be empty meaning corresponding source has no attachments.
            This way each source can have different attachment files
          * total directories differ from total source files
            return MKVCommandParser.attachmentsString for every Source
    """

    def __init__(self, strCommand=None):

        # for iterator
        self.__index = 0
        self.strCommand = strCommand

    def _initVars(self):

        self.__attachmentMatch = None
        self.__cmdLineAttachmentsStr = None
        self.__totalSourceFiles = None
        self.__strCommand = None

        self.__attachments = []
        self.__attachmentsFiles = []
        self.__attachmentsDirs = []
        self.__attachmentsStr = []
        self.__attachmentsDirByEpisode = False

        self.__cmdLineAttachments = []
        self.__cmdLineAttachmentsFiles = []
        self.__cmdLineAttachmentsDirs = []

    def __bool__(self):
        return bool(self.__attachments)

    def __contains__(self, item):
        return item in self.__cmdLineAttachments

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.__cmdLineAttachments)

    def __getitem__(self, index):
        return self.__cmdLineAttachments[index]

    def __next__(self):
        if self.__index >= len(self.__cmdLineAttachments):
            self.__index = 0
            raise StopIteration
        else:
            self.__index += 1

    @property
    def attachments(self):
        return self.__attachments

    @property
    def attachmentsDirs(self):
        return self.__attachmentsDirs

    @property
    def isAttachmentsDirByEpisode(self):
        return self.__attachmentsDirByEpisode

    @property
    def attachmentsFiles(self):
        return self.__attachmentsFiles

    @property
    def attachmentsStr(self):
        return self.__attachmentsStr

    @property
    def cmdLineAttachments(self):
        return self.__cmdLineAttachments

    @property
    def cmdLineAttachmentsDirs(self):
        return self.__cmdLineAttachmentsDirs

    @property
    def cmdLineAttachmentsFiles(self):
        return self.__cmdLineAttachmentsFiles

    @property
    def attachmentsMatchString(self):
        strTmp = None
        if self.__cmdLineAttachments:
            span = self.attachmentsSpan
            strTmp = self.strCommand[span[0] : span[1]]
        return strTmp

    @property
    def attachmentsSpan(self):
        span = ()
        if self.__cmdLineAttachments:
            span = (
                self.__cmdLineAttachments[0].span[0],
                self.__cmdLineAttachments[-1].span[1],
            )
        return span

    @property
    def strCommand(self):
        return self.__strCommand

    @strCommand.setter
    def strCommand(self, value):
        if isinstance(value, str):
            self._initVars()
            self.__strCommand = value
            self._parse()
            if self.__cmdLineAttachments:
                self._readDirs()

    def _parse(self):

        reSourceEx = re.compile(r"'\('\s(.*?)\s'\)'")

        reAttachmentsEx = re.compile(
            (
                r"--attachment-name (.*?) --attachment-mime-type "
                r"(.*?) --attach-file (.*?)(?= --)"
            )
        )

        if match := reSourceEx.search(self.__strCommand):

            f = unQuote(match.group(1))
            p = Path(f)

            try:
                test = p.is_file()
            except OSError:
                pass
            else:
                if test:
                    fid = fid = [
                        x for x in p.parent.glob("*" + p.suffix) if x.is_file()
                    ]
                    self.__totalSourceFiles = len(fid)

        if matchAttachments := reAttachmentsEx.finditer(self.__strCommand):
            dirs = set()
            for match in matchAttachments:
                attachment = MKVAttachment(match.groups(), match.span(), match.group())
                self.__cmdLineAttachments.append(attachment)
                p = Path(unQuote(match.group(3)))
                try:
                    test = p.is_file()
                except OSError:
                    pass
                else:
                    if test:
                        self.__cmdLineAttachmentsFiles.append(p)
                        dirs.add(p.parent)
            self.__cmdLineAttachmentsDirs.extend(list(dirs))

    def _readDirs(self):

        if len(self.__cmdLineAttachmentsDirs) == 1:
            # Check parent for directories
            d = self.__cmdLineAttachmentsDirs[0]
            pd = self.__cmdLineAttachmentsDirs[0].parent
            did = [x for x in pd.glob("*") if x.is_dir()]
            fid = [x for x in d.glob("*") if x.is_file()]

            if self.__totalSourceFiles == len(did):
                self.__attachmentsDirByEpisode = True
                self.__attachmentsDirs.extend(did)
                for d in did:
                    fid = [x for x in d.glob("*") if x.is_file()]
                    lstTmp = []
                    lstTmp.extend(fid)
                    self.__attachmentsFiles.append(lstTmp)
                    self.__attachmentsStr.append(attachmentsToStr(lstTmp))
            else:
                self.__attachmentsDirs.extend(
                    self.__cmdLineAttachmentsDirs * self.__totalSourceFiles
                )
                self.__attachmentsStr.extend(
                    [self.attachmentsMatchString] * self.__totalSourceFiles
                )

        else:
            self.__attachmentsStr.extend(
                [self.attachmentsMatchString] * self.__totalSourceFiles
            )


def _attachmentToStr(attachFile):

    strTmp = "--attachment-name " + shlex.quote(attachFile.name)
    strTmp += " --attachment-mime-type " + mimeType(attachFile)
    strTmp += " --attach-file " + shlex.quote(str(attachFile))

    return strTmp


def attachmentsToStr(attachFiles):
    """
    attachmentsToStr convert list of attachment files to mkvmerge option

    Args:
        attachFiles (list): list of attachment files

    Returns:
        str: mkvmerge options string
    """

    strTmp = ""
    bFirst = True
    for f in attachFiles:
        if bFirst:
            strTmp = _attachmentToStr(f)
            bFirst = False
        else:
            strTmp += " " + _attachmentToStr(f)

    return strTmp


def mimeType(fileName):
    """
    mimeType return mime type of known files by suffix

    Args:
        fileName (filepath.Path): file Path object

    Returns:
        str: known mime type for file defaults to application/octet-stream
    """

    if fileName.suffix.upper() in [".TTF", ".TTC"]:
        return "application/x-truetype-font"
    elif fileName.suffix.upper() in [".OTF"]:
        return "application/vnd.ms-opentype"
    elif fileName.suffix.upper() in [".ZIP"]:
        return "application/zip"
    elif fileName.suffix.upper() in [".JPG", ".JPEG"]:
        return "image/jpeg"
    elif fileName.suffix.upper() in [".PNG"]:
        return "image/png"

    return "application/octet-stream"



"""
This class is used to construct the list used to execute the MKVMerge
command line

strCommand = Command obtained from mkvtoolnix-gui with the modifications
    needed for Multiplexing a series in a directory

path for executable and target options are parsed from the command line

"""

import re
import shlex
import logging

from pathlib import Path

from vsutillib.files import stripEncaseQuotes
from vsutillib.misc import XLate

from .verifycommand import VerifyMKVCommand

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class MKVCommand(object):
    """
    Class to work with **mkvmerge** command part of MKVToolnix_

    Args:
        strCommand (:obj:`str`, optional): command line as generated by mkvtoolnix-gui
    """

    __log = False

    @classmethod
    def classLog(cls, setLogging=None):
        """
        get/set logging at class level
        every class instance will log
        unless overwritten

        Args:
            setLogging (bool):
                - True class will log
                - False turn off logging
                - None returns current Value

        Returns:
            bool:

            returns the current value set
        """

        if setLogging is not None:
            if isinstance(setLogging, bool):
                cls.__log = setLogging

        return cls.__log

    def __init__(self, strCommand=None, log=None):

        # self.__destinationDirectory = None
        self.__lstCommands = []
        self.__strShellCommand = None
        self.__strError = ""
        self.__bErrorFound = False
        self.__workFiles = _WorkFiles()
        self.__commandTemplate = None
        self.__filesInDirsByKey = None

        self.__log = None
        self.log = log

        # for iterator
        self.__index = 0

        if strCommand is not None:
            self._initHelper(strCommand)

    def _initHelper(self, strCommand, bRemoveTitle=True): # pylint: disable=unused-argument

        if strCommand is None:
            self.__reset()
            verify = False
        else:
            verify = VerifyMKVCommand(strCommand, log=self.log)

        if verify:

            self.__strShellCommand = strCommand
            self.__bErrorFound = False
            lstSources = []
            bHasChaptersFile = False

            strCommand = verify.bashCommand

            #reOutputFile = r"\-\-output\s(.*?)\s\-\-"
            #reChaptersFile = r"\-\-chapters\s(.*?)\s\-\-"
            #reSourceFile = r"('\('\s(.*?)\s'\)')"

            reOutputFileEx = re.compile(r"--output\s(.*?)\s--")
            reChaptersFileEx = re.compile(r"--chapters\s(.*?)\s--")
            reSourcesEx = re.compile(r"('\('\s(.*?)\s'\)')")

            # search for the output file
            if match := reOutputFileEx.search(strCommand):
                strOutputFile = match.group(1)

            outputFile = verify.outputFile

            # search for chapters file
            # match = reChaptersFileEx.search(strCommand)
            if match := reChaptersFileEx.search(strCommand):
                strChaptersFile = match.group(1)
                chaptersFile = verify.chaptersFile
                bHasChaptersFile = True

            # search for the source files
            # this have to exist in the
            # file system
            if match := reSourcesEx.findall((strCommand)):
                for m in match:
                    lstSources.append(m)

            newCommandTemplate = strCommand

            filesInDirsByKey = {}
            lstBaseFiles = []

            lenOfListOfFiles = 0

            for index, source in enumerate(lstSources):

                # Set source files
                sub, fileName = source
                key = "<SOURCE{}>".format(str(index))
                fileName = fileName.replace(r"'\''", "'")
                f = Path(stripEncaseQuotes(fileName))
                d = f.parent
                fd = [x for x in d.glob("*" + f.suffix) if x.is_file()]
                fd.sort(key=_strPath)

                # Check all lists have same number of files
                if lenOfListOfFiles == 0:
                    lenOfListOfFiles = len(fd)

                elif lenOfListOfFiles != len(fd):
                    self.__reset()
                    self.__bErrorFound = True
                    self.__strError = "List of files are not equal."

                # Set output files
                if index == 0:
                    od = []
                    for o in fd:
                        of = outputFile.parent.joinpath(o.stem + ".mkv")
                        of = _resolveOverwrite(of)
                        od.append(of)
                    filesInDirsByKey[_Key.outputFile] = od
                    newCommandTemplate = newCommandTemplate.replace(
                        strOutputFile, _Key.outputFile, 1
                    )

                lstBaseFiles.append(f)  # backwards compatible
                filesInDirsByKey[key] = fd
                newCommandTemplate = newCommandTemplate.replace(sub, key, 1)

            # Set output files
            if bHasChaptersFile:
                d = chaptersFile.parent
                cd = [x for x in d.glob("*" + chaptersFile.suffix) if x.is_file()]
                cd.sort(key=_strPath)

                if lenOfListOfFiles != len(cd):
                    self.__reset()
                    self.__bErrorFound = True
                    self.__strError = "List of files are not equal."

                filesInDirsByKey[_Key.chaptersFile] = cd

                newCommandTemplate = newCommandTemplate.replace(
                    strChaptersFile, _Key.chaptersFile, 1
                )

            if not self.__bErrorFound:
                self.__commandTemplate = newCommandTemplate
                self.__filesInDirsByKey = filesInDirsByKey

                self.__workFiles.baseFiles = lstBaseFiles
                self.__workFiles.chaptersFiles = (
                    None
                    if _Key.chaptersFile not in filesInDirsByKey
                    else filesInDirsByKey[_Key.chaptersFile]
                )

                self._generateCommands()
            else:
                if self.log:
                    MODULELOG.error("MKV0005: %s", self.__strError)

        else:
            # error cannot process command
            self.__reset()
            self.__strError = "Error parsing command line."
            self.__strShellCommand = strCommand
            self.__bErrorFound = True

            if self.log:
                MODULELOG.error("MKV0005: %s", self.__strError)

    def __reset(self):
        """Reset variable properties"""

        # self.__destinationDirectory = None
        self.__lstCommands = []
        self.__strShellCommand = None
        self.__strError = ""
        self.__index = 0
        self.__workFiles.clear()
        self.__commandTemplate = None
        self.__filesInDirsByKey = None

    def __bool__(self):
        return not self.__bErrorFound

    def __contains__(self, item):
        return item in self.__lstCommands

    def __getitem__(self, index):
        return (
            self.__lstCommands[index],
            self.__workFiles.baseFiles,
            self.__workFiles.sourceFiles[index],
            self.__workFiles.destinationFiles[index],
            None
            if not self.__workFiles.chaptersFiles
            else self.__workFiles.chaptersFiles[index],
        )

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.__lstCommands)

    def __next__(self):
        if self.__index >= len(self.__lstCommands):
            self.__index = 0
            raise StopIteration
        else:
            self.__index += 1
            return self.__getitem__(self.__index - 1)

            # return [self.__lstCommands[self.__index - 1],
            #        self.__workFiles.baseFiles,
            #        self.__workFiles.sourceFiles[self.__index - 1],
            #        self.__workFiles.destinationFiles[self.__index - 1]]

    def _generateCommands(self, bRemoveTitle=True):

        # lstCommandsNew = []

        newCommandNew = self.__commandTemplate

        totalCommands = len(self.__filesInDirsByKey[_Key.outputFile])
        lstSourceFilesNew = []

        self.__lstCommands = []

        for i in range(totalCommands):

            e = {}
            s = []

            for k, v in self.__filesInDirsByKey.items():

                e[k] = shlex.quote(
                    str(v[i])
                )  # add the filenames as shlex.quote strings form pathlib.Path

                if k != _Key.outputFile:
                    s.append(v[i])  # save source files as pathlib.Path

            xLate = XLate(e)  # instantiate regex dictionary translator
            lstSourceFilesNew.append(s)  # Save Source Files in list

            shellCommand = shlex.split(
                xLate.xLate(newCommandNew)
            )  # save command as shlex.split to submit to Pipe

            if bRemoveTitle and shellCommand:
                # Remove title if found since this is for batch processing
                # the title will propagate to all the files maybe erroneously.
                # This field is preserved from the source files.

                while "--title" in shellCommand:

                    i = shellCommand.index("--title")
                    del shellCommand[i : i + 2]

            self.__lstCommands.append(shellCommand)  # save command in list

        self.__workFiles.sourceFiles = lstSourceFilesNew  # redundant for rename
        self.__workFiles.destinationFiles = self.__filesInDirsByKey[_Key.outputFile]

        if self.log:
            MODULELOG.debug("MKV0001: Command template %s", str(self.__commandTemplate))
            MODULELOG.debug("MKV0002: Base files %s", str(self.__workFiles.baseFiles))
            MODULELOG.debug(
                "MKV0003: Source files %s", str(self.__workFiles.sourceFiles)
            )
            MODULELOG.debug(
                "MKV0004: Destination files %s", str(self.__workFiles.destinationFiles)
            )

    @property
    def log(self):
        """
        class property can be used to override the class global
        logging setting

            bool:

            True if logging is enable False otherwise
        """
        if self.__log is not None:
            return self.__log

        return MKVCommand.classLog()

    @log.setter
    def log(self, value):
        """set instance log variable"""
        if isinstance(value, bool) or value is None:
            self.__log = value

    @property
    def command(self):
        """
        get/set command produced by mkvtoolnix-gui

        Returns:
            str:

            original command set
        """
        return self.__strShellCommand

    @command.setter
    def command(self, value):
        """Update command through property"""
        if isinstance(value, str):
            self.__reset()
            self._initHelper(value, bRemoveTitle=True)

    @property
    def baseFiles(self):
        """
        files parsed from command

        Returns:
            list:

            list with the files as parsed from command
        """
        return self.__workFiles.baseFiles

    @property
    def sourceFiles(self):
        """
        source files read from respective directories

        Returns:
            list:

            list with all the source files
        """
        return self.__workFiles.sourceFiles

    @property
    def destinationFiles(self):
        """
        destination files

        Returns:
            list:

            list with destination files
        """
        return self.__workFiles.destinationFiles

    @property
    def template(self):
        """
        template to construct the commands

        Returns:
            list:

            command template
        """
        return self.__commandTemplate

    @property
    def error(self):
        """
        message of any error found during evaluation of command

        Returns:
            str:

            error description
        """
        return self.__strError

    def renameOutputFiles(self, fileNames):
        """
        Mothod for renaming the output files

        Arguments:
            fileNames {list} -- list of file names
        """

        totalNames = len(self.__filesInDirsByKey[_Key.outputFile])
        totalRenames = len(fileNames)

        if totalNames != totalRenames:
            self.__strError = "Files to rename and new names not equal length."

        else:
            self.__filesInDirsByKey[_Key.outputFile] = fileNames
            self._generateCommands()


class _WorkFiles:
    """Files read from directories"""

    def __init__(self):

        self.baseFiles = []
        self.sourceFiles = []
        self.destinationFiles = []
        self.chaptersFiles = []

    def clear(self):
        """Clear file lists"""

        self.baseFiles = []
        self.sourceFiles = []
        self.destinationFiles = []
        self.chaptersFiles = []


class _Key:

    outputFile = "<OUTPUTFILE>"
    chaptersFile = "<CHAPTERS>"


def _resolveOverwrite(fileName, strPrefix="new-"):

    fileNameTmp = fileName

    # Check if destination file exist and add prefix if it does
    if fileNameTmp.is_file():

        strSuffix = ""
        n = 1

        while True:
            fileNameTmp = fileNameTmp.parent.joinpath(
                strPrefix + fileName.stem + strSuffix + fileName.suffix
            )

            if fileNameTmp.is_file():
                strSuffix = " ({})".format(n)
                n += 1
            else:
                break

    # video-S01E01.mkv
    return fileNameTmp


def _strPath(value):
    """
    Convenience function to help sort

    Arguments:
        value {various types} -- value to convert to string

    Returns:
        str -- argument received converted to string
    """
    return str(value)



"""
classes related to files
========================

ConfigurationSettings - read/save dictionary of terms
    from file, keys must be strings

"""

from .configurationsettings import ConfigurationSettings
from .DisplayPath import DisplayPath



"""
Helper classes for QThreadPool
"""


import logging
import traceback
import sys

from PySide2.QtCore import QObject, QRunnable, Signal, Slot


MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class WorkerSignals(QObject):
    '''
    Defines the signals available from a running worker thread.

    Supported signals are:

    finished
        No data

    error
        `tuple` (exectype, value, traceback.format_exc() )

    result
        `object` data returned from processing, anything
    '''
    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)

class Worker(QRunnable):
    '''
    Worker thread

    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.

    :param function: The function callback to run on this worker thread. Supplied args and
                     kwargs will be passed through to the runner.
    :type function: function
    :param args: Arguments to pass to the callback function
    :param kwargs: Keywords to pass to the callback function
    '''

    log = False

    def __init__(self, function, *args, **kwargs):
        super().__init__()

        # Store constructor arguments (re-used for processing)
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

    @Slot()
    def run(self):
        '''
        Initialise the runner function with passed args, kwargs.
        '''

        # Retrieve args/kwargs here; and fire processing using them
        # pylint: disable-msg=W0702
        # have to capture all exceptions using sys.exc_info()
        # to sort out what is happening
        try:
            result = self.function(*self.args, **self.kwargs)
        except:
            traceback.print_exc()
            excepttype, value = sys.exc_info()[:2]
            self.signals.error.emit((excepttype, value, traceback.format_exc()))
        else:
            self.signals.result.emit(result)  # Return the result of the processing
        finally:
            self.signals.finished.emit()  # Done



"""
Network related utilities
"""

import json
import socket
import urllib.parse
import urllib.request

REMOTE_SERVER = "www.google.com"

def isConnected(hostname=None):
    """
    Convenience function to test for web server connection
    no parameters checks for internet connection

    Args:
        hostname (str): hostname to test for connection
    """

    if hostname is None:
        hostname = REMOTE_SERVER

    try:
        # see if we can resolve the host name -- tells us if there is
        # a DNS listening
        host = socket.gethostbyname(hostname)

        # connect to the host -- tells us if the host is actually
        # reachable
        s = socket.create_connection((host, 80), 2)
        s.close()

        return True

    except OSError:

        pass

    return False


def urlSearchJson(url, data=None, headers=None):
    """
    url request to a site that returns json data

    Args:
        url (str): URL for request
        data (:obj:`dict`, optional): dictionary with data for GET
        headers (:obj:`dict`, optional): dictionary with any needed headers

    Returns:
        dict:

        dictionary with the json response
    """

    if data is not None:
        request = urllib.request.Request(url, data=data)
    else:
        request = urllib.request.Request(url)

    if headers is not None:
        for key in headers:
            request.add_header(key, headers[key])

    result = urllib.request.urlopen(request)

    if result.status == 200:

        bJson = result.read()
        data = json.loads(bJson.decode('utf-8'))
        return data

    return None



"""
Multithreading Class base on threading.Thread
"""

import logging
import sys
import threading
import traceback

MODULELOG = logging.getLogger(__name__)
MODULELOG.addHandler(logging.NullHandler())


class GenericThreadWorker(threading.Thread):
    """
    Generic Thread worker

    Args:
        function (function): function to submit to Process
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    log = False

    def __init__(self, function, *args, **kwargs):
        super(GenericThreadWorker, self).__init__()

        self.function = function
        self.args = args
        self.kwargs = kwargs

    def run(self):
        """
        Override run initialise and starts the worker function
        with passed args, kwargs.
        """

        try:
            self.function(*self.args, **self.kwargs)
        except:  # pylint: disable=bare-except
            traceback.print_exc()

        return


class QueueThreadWorker(threading.Thread):
    """
    Generic Queue process Thread worker

    Args:
        queue (Queue): Queue to process
        function (function): function to submit to Process
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    log = False

    def __init__(self, queue, function, *args, **kwargs):
        super(QueueThreadWorker, self).__init__()

        self.queue = queue
        self.function = function
        self.args = args
        self.kwargs = kwargs

    def run(self):
        """
        Override run gets next job from queue. The initialise and
        starts the worker function with passed args, kwargs and
        nextJob.
        """

        while True:
            # Get the work from the queue and expand the tuple
            nextJob = self.queue.get()
            try:
                self.function(nextJob, *self.args, **self.kwargs)
            finally:
                self.queue.task_done()


class ThreadWorker(threading.Thread):
    """
    Worker thread

    Inherits from threading.Tread to handle worker thread setup, signals and wrap-up.

    Args:
        function (function): Function to submit to Thread.
        funcFinished (function): Call back function when thread finishes.
        funcError (function): Call back function when an error occurs.
        funcResult (function): Call back function with the result of the execution.
        funcStart (function): Call back function when an Thread is about to start.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    log = False

    def __init__(self,
                 function,
                 *args,
                 **kwargs):
        super(ThreadWorker, self).__init__()

        # Store constructor arguments (re-used for processing)
        self.function = function
        self.args = args

        # remove optional known named parameters
        self.funcStart = kwargs.pop('funcStart', None)
        self.funcFinished = kwargs.pop('funcFinished', None)
        self.funcError = kwargs.pop('funcError', None)
        self.funcResult = kwargs.pop('funcResult', None)

        self.kwargs = kwargs # pass rest of named parameters to self.function

    def run(self):
        """
        Override run initialise and starts the worker function
        with passed args, kwargs.
        """

        if callable(self.funcStart):
            self.funcStart()  # Done

        # Retrieve args/kwargs here; and fire processing using them
        # pylint: disable-msg=W0702
        # have to capture all exceptions using sys.exc_info()
        # to sort out what is happening
        try:
            result = self.function(*self.args, **self.kwargs)
        except:
            traceback.print_exc()
            excepttype, value = sys.exc_info()[:2]
            if callable(self.funcError):
                self.funcError((excepttype, value, traceback.format_exc()))
        else:
            if callable(self.funcResult):
                self.funcResult(result)  # Return the result of the processing
        finally:
            if callable(self.funcFinished):
                self.funcFinished()  # Done



"""
vsutillib version and configuration constants

config can be use by other applications

import vsutillib.config as config

config.init(configFile='appName-config.xml',
            filesRoot='.appName',
            logFile='file.log',
            name='appName',
            version='1.0')

and

config.close()

config.data is files.ConfigurationSettings class instance

before application exit
"""

import logging
import sys
from pathlib import Path

from vsutillib.files import ConfigurationSettings
from vsutillib.log import LogRotateFileHandler


__VERSION = (1, 6, "3")


FILES_VERSION = "1.6.3"
LOG_VERSION = "1.6.0"
MACOS_VERSION = "1.6.0"
MEDIA_VERSION = "1.6.2"
MISC_VERSION = "1.6.0"
MKV_VERSION = "1.6.2"
NETWORK_VERSION = "1.6.0"
PROCESS_VERSION = "1.6.0"
PYQT_VERSION = "1.6.2"
SCRIPTS_VERSION = "1.6.2"
SQL_VERSION = "1.6.2"
VSXML_VERSION = "1.6.0"

LXML_VERSION = ">=4.5"
NATSORT_VERSION = ">=7.0.1"
PYMEDIAINFO_VERSION = ">=4.2.1"
PYSIDE2_VERSION = ">=5.14"
PYTHON_VERSION = ">=3.8.1, <3.9"
PYTHONVERSIONS = ">=3.8.1, <3.9"

APPNAME = "vsutillib"
VERSION = ".".join(map(str, __VERSION))
AUTHOR = "Efrain Vergara"
EMAIL = "akai10tsuki@gmail.com"
PACKAGES = [APPNAME]

COPYRIGHT = "2018-2020, Efrain Vergara"
LICENSE = "MIT"
DESCRIPTION = "Library module with miscellaneous convenience functions and classes"
NAME = "vsutillib"
KEYWORDS = "mkv multimedia video audio configuration"
REQUIRED = [
    "lxml" + LXML_VERSION,
    "pymediainfo" + PYMEDIAINFO_VERSION,
    "PySide2" + PYSIDE2_VERSION,
    "natsort" + NATSORT_VERSION,
]
URL = "https://github.com/akai10tsuki/vsutillib"
PYPI = "https://pypi.org/project/vsutillib/"
PROJECTURLS = {
    "Source": "https://pypi.org/project/vsutillib/#files",
}

CONFIGFILE = "config.xml"
FILESROOT = "." + APPNAME
LOGFILE = APPNAME + ".log"
IMAGEFILESPATH = ""

__version__ = VERSION

data = ConfigurationSettings()  # pylint: disable=invalid-name


def init(filesRoot=None, configFile=None, logFile=None, name=None, version=None):
    """
    configures the system to save application configuration to xml file

    Args:
        filesRoot (str, optional): root folder on ~ for files. Defaults to [.vsutillib].
        configFile (str, optional): name of configuration file. Defaults to [config.xml].
        logFile (str, optional): name of logging file. Defaults to [vsutillib.log].
        name (str, optional): name of application. Defaults to [vsutillib].
        version (str, optional): appplication version . Defaults to [vsutillib version].
    """

    if filesRoot is None:
        filesPath = Path(Path.home(), FILESROOT)
    else:
        filesPath = Path(Path.home(), filesRoot)

    filesPath.mkdir(parents=True, exist_ok=True)

    if configFile is None:
        cfgFile = Path(filesPath, CONFIGFILE)
    else:
        cfgFile = Path(filesPath, configFile)

    data.setConfigFile(cfgFile)
    data.readFromFile()

    if logFile is None:
        loggingFile = Path(filesPath, LOGFILE)
    else:
        loggingFile = Path(filesPath, logFile)

    loghandler = LogRotateFileHandler(loggingFile, backupCount=10)
    formatter = logging.Formatter("%(asctime)s %(levelname)-8s %(name)s %(message)s")
    loghandler.setFormatter(formatter)

    logging.getLogger("").setLevel(logging.DEBUG)
    logging.getLogger("").addHandler(loghandler)
    logging.info("App Start.")
    logging.info("Python: %s", sys.version)

    if name is None:
        appName = NAME
    else:
        appName = name

    if version is None:
        appVersion = VERSION
    else:
        appVersion = version

    logging.info("%s-%s", appName, appVersion)


def close():
    """
    save configuration to file and write end message to log file
    """
    data.saveToFile()

    logging.info("App End.")



"""
subclass of QPushButton to save text, shortcut and tooltip
this information is used for internationalization
"""

from PySide2.QtWidgets import QPushButton


class QPushButtonWidget(QPushButton):
    """
    QPushButton subclass of QAction save original shortcut and tooltip for locale application

    Args:
        shortcut (str, optional): original shortcut string representation. Defaults to None.
        tooltip (str, optional): original tooltip. Defaults to None.
    """

    def __init__(self, *args, function=None, toolTip=None, **kwargs):
        super().__init__(*args, **kwargs)

        for p in args:
            if isinstance(p, str):
                # Save first string found assume is the button label
                self.originalText = p
                break

        self.toolTip = toolTip

        if function is not None:
            self.clicked.connect(function)

        if toolTip is not None:
            self.setToolTip(toolTip)

    def setToolTip(self, toolTip, *args, **kwargs):

        if self.toolTip is None:
            self.toolTip = toolTip

        super().setToolTip(toolTip, *args, **kwargs)
